# =============================================================================
# Minet CLI Utils
# =============================================================================
#
# Miscellaneous helpers used by the CLI tools.
#
import csv
import sys
import argparse
from collections import namedtuple
from tqdm import tqdm


def print_err(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def safe_index(l, e):
    try:
        return l.index(e)
    except ValueError:
        return None


def custom_reader(f, target_header):

    reader = csv.reader(f)

    headers = next(reader, None)

    if isinstance(target_header, tuple):
        HeaderPositions = namedtuple('HeaderPositions', target_header)
        position = HeaderPositions(**{t: safe_index(headers, t) for t in target_header})
    else:
        position = safe_index(headers, target_header)

    return headers, position, reader


class DummyTqdmFile(object):
    """
    Dummy file-like that will write to tqdm. Taken straight from the lib's
    documentation: https://github.com/tqdm/tqdm but modified for minet use
    case regarding stdout piping.
    """
    file = None

    def __init__(self, file):
        self.file = file

    def write(self, x):
        # Avoid print() second call (useless \n)
        tqdm.write(x, file=self.file, end='')

    def flush(self):
        return self.file.flush()

    def close(self):
        pass


class BooleanAction(argparse.Action):
    """
    Custom argparse action to handle --no-* flags.
    Taken from: https://thisdataguy.com/2017/07/03/no-options-with-argparse-and-python/
    """
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        super(BooleanAction, self).__init__(option_strings, dest, nargs=0, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, False if option_string.startswith('--no') else True)



# =============================================================================
# Minet Scrape CLI Action
# =============================================================================
#
# Logic of the scrape action.
#
import csv
import json
import sys
import codecs
from os.path import join
from multiprocessing import Pool
from tqdm import tqdm

from minet.scrape import scrape, headers_from_definition
from minet.cli.utils import custom_reader, DummyTqdmFile

ERROR_REPORTERS = {
    UnicodeDecodeError: 'wrong-encoding'
}


def worker(payload):
    line, path, encoding, content, scraper = payload

    # Reading from file
    if content is None:
        with codecs.open(path, 'r', encoding=encoding, errors='replace') as f:
            try:
                content = f.read()
            except UnicodeDecodeError as e:
                return e, line, None

    # Attempting to scrape
    items = list(scrape(content, scraper))

    return None, line, items


def scrape_action(namespace):
    input_headers, pos, reader = custom_reader(namespace.report, ('status', 'filename', 'encoding', 'raw_content'))

    if namespace.output is None:
        output_file = DummyTqdmFile(sys.stdout)
    else:
        output_file = open(namespace.output, 'w')

    # Parsing scraper definition
    scraper = json.load(namespace.scraper)

    output_headers = headers_from_definition(scraper)
    output_writer = csv.DictWriter(output_file, fieldnames=output_headers)
    output_writer.writeheader()

    def files():
        for line in reader:
            status = int(line[pos.status]) if line[pos.status] else None

            if status is None or status >= 400:
                loading_bar.update()
                continue

            if pos.raw_content is not None:
                yield line, None, None, line[pos.raw_content], scraper
                continue

            path = join(namespace.input_directory, line[pos.filename])
            encoding = line[pos.encoding].strip() or 'utf-8'

            yield line, path, encoding, None, scraper

    loading_bar = tqdm(
        desc='Scraping pages',
        total=namespace.total,
        dynamic_ncols=True,
        unit=' pages'
    )

    with Pool(namespace.processes) as pool:
        for error, line, items in pool.imap_unordered(worker, files()):
            loading_bar.update()

            for item in items:
                if not isinstance(item, dict):
                    item = {'value': item}

                output_writer.writerow(item)

    output_file.close()



# =============================================================================
# Minet CrowdTangle Search CLI Action
# =============================================================================
#
# Logic of the `ct search` action.
#
import json
from urllib.parse import quote

from minet.cli.crowdtangle.utils import create_paginated_action
from minet.cli.crowdtangle.posts import CSV_HEADERS, format_post_for_csv

URL_TEMPLATE = 'https://api.crowdtangle.com/posts/search?count=100&sortBy=%(sort_by)s&token=%(token)s&searchTerm=%(terms)s'


def forge_posts_url(namespace):
    base_url = URL_TEMPLATE % {
        'sort_by': namespace.sort_by,
        'token': namespace.token,
        'terms': quote(namespace.terms)
    }

    if namespace.start_date:
        base_url += '&startDate=%s' % namespace.start_date

    if namespace.end_date:
        base_url += '&endDate=%s' % namespace.end_date

    if namespace.platforms:
        base_url += '&platforms=%s' % namespace.platforms

    if namespace.types:
        base_url += '&types=%s' % namespace.types

    if namespace.offset:
        base_url += '&offset=%s' % namespace.offset

    return base_url


crowdtangle_search_action = create_paginated_action(
    url_forge=forge_posts_url,
    csv_headers=CSV_HEADERS,
    csv_formatter=format_post_for_csv,
    item_name='posts',
    item_key='posts'
)



# =============================================================================
# Minet CrowdTangle CLI Defaults
# =============================================================================
#
# General constants used throughout the CrowdTangle actions.
#

CROWDTANGLE_DEFAULT_WAIT_TIME = 10

CROWDTANGLE_POST_TYPES = [
    'episode',
    'extra_clip',
    'link',
    'live_video',
    'live_video_complete',
    'live_video_scheduled',
    'native_video',
    'photo',
    'status',
    'trailer',
    'tweet',
    'video',
    'vine',
    'youtube'
]

CROWDTANGLE_SORT_TYPES = [
    'date',
    'interaction_rate',
    'overperforming',
    'total_interactions',
    'underperforming'
]



# =============================================================================
# Minet CrowdTangle Posts CLI Action
# =============================================================================
#
# Logic of the `ct posts` action.
#
import json

from minet.cli.crowdtangle.utils import create_paginated_action

URL_TEMPLATE = 'https://api.crowdtangle.com/posts?count=100&sortBy=%(sort_by)s&token=%(token)s'


def forge_posts_url(namespace):
    base_url = URL_TEMPLATE % {
        'sort_by': namespace.sort_by,
        'token': namespace.token
    }

    if namespace.language:
        base_url += '&language=%s' % namespace.language

    if namespace.start_date:
        base_url += '&startDate=%s' % namespace.start_date

    if namespace.end_date:
        base_url += '&endDate=%s' % namespace.end_date

    if namespace.list_ids:
        base_url += '&list-ids=%s' % namespace.list_ids

    return base_url

CSV_HEADERS = [
    'ct_id',
    'id',
    'platform',
    'type',
    'title',
    'caption',
    'message',
    'description',
    'date',
    'updated',
    'link',
    'post_url',
    'score',
    'video_length_ms',
    'live_video_status'
]

STATISTICS = [
    'like',
    'share',
    'favorite',
    'comment',
    'love',
    'wow',
    'haha',
    'sad',
    'angry',
    'thankful'
]

for name in STATISTICS:
    CSV_HEADERS.append('actual_%s_count' % name)
    CSV_HEADERS.append('expected_%s_count' % name)

CSV_HEADERS = CSV_HEADERS + [
    'account_ct_id',
    'account_id',
    'account_platform',
    'account_name',
    'account_handle',
    'account_profile_image',
    'account_subscriber_count',
    'account_url',
    'account_verified',
    'links',
    'expanded_links',
    'media'
]


def format_post_for_csv(namespace, post):
    row = [
        post['id'],
        post['platformId'],
        post['platform'],
        post['type'],
        post.get('title', ''),
        post.get('caption', ''),
        post.get('message', ''),
        post.get('description', ''),
        post['date'],
        post['updated'],
        post.get('link', ''),
        post['postUrl'],
        post['score'],
        post.get('videoLengthMS', ''),
        post.get('liveVideoStatus', '')
    ]

    stats = post['statistics']
    actual_stats = stats['actual']
    expected_stats = stats['expected']

    for name in STATISTICS:
        key = '%sCount' % name

        row.append(actual_stats.get(key, ''))
        row.append(expected_stats.get(key, ''))

    account = post['account']

    links = ''
    expanded_links = ''

    if 'expandedLinks' in post:
        links = '|'.join(link['original'] for link in post['expandedLinks'])
        expanded_links = '|'.join(link['expanded'] for link in post['expandedLinks'])

    row.extend([
        account['id'],
        account.get('platformId', ''),
        account['platform'],
        account['name'],
        account.get('handle', ''),
        account['profileImage'],
        account['subscriberCount'],
        account['url'],
        '1' if account['verified'] else '',
        links,
        expanded_links,
        json.dumps(post['media'], ensure_ascii=False) if 'media' in post else ''
    ])

    return row

crowdtangle_posts_action = create_paginated_action(
    url_forge=forge_posts_url,
    csv_headers=CSV_HEADERS,
    csv_formatter=format_post_for_csv,
    item_name='posts',
    item_key='posts'
)



# =============================================================================
# Minet CrowdTangle Posts CLI Action
# =============================================================================
#
# Logic of the `ct posts` action.
#
import csv
import sys
import json
import urllib3
import certifi

from minet.cli.utils import print_err

URL_TEMPLATE = 'https://api.crowdtangle.com/lists?token=%s'

CSV_HEADERS = [
    'id',
    'title',
    'type'
]


def format_list_for_csv(l):
    return [
        l['id'],
        l['title'],
        l['type']
    ]


def crowdtangle_lists_action(namespace, output_file):
    http = urllib3.PoolManager(
        cert_reqs='CERT_REQUIRED',
        ca_certs=certifi.where(),
    )

    result = http.request('GET', URL_TEMPLATE % namespace.token)

    if result.status == 401:
        print_err('Your API token is invalid.')
        print_err('Check that you indicated a valid one using the `--token` argument.')
        sys.exit(1)

    if result.status >= 400:
        print_err(result.data, result.status)
        sys.exit(1)

    writer = csv.writer(output_file)
    writer.writerow(CSV_HEADERS)

    data = json.loads(result.data)

    for l in data['result']['lists']:
        writer.writerow(format_list_for_csv(l))



# =============================================================================
# Minet Extract Content CLI Action
# =============================================================================
#
# Logic of the extract action.
#
import csv
import sys
import codecs
import warnings
from os.path import join
from multiprocessing import Pool
from tqdm import tqdm
from dragnet import extract_content

from minet.cli.utils import custom_reader, DummyTqdmFile

OUTPUT_ADDITIONAL_HEADERS = ['extract_error', 'extracted_text']

ERROR_REPORTERS = {
    UnicodeDecodeError: 'wrong-encoding'
}


def worker(payload):

    line, path, encoding = payload

    # Reading file
    with codecs.open(path, 'r', encoding=encoding, errors='replace') as f:
        try:
            raw_html = f.read()
        except UnicodeDecodeError as e:
            return e, line, None

    # Attempting extraction
    try:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            content = extract_content(raw_html)
    except BaseException as e:
        # TODO: I don't know yet what can happen
        raise

    return None, line, content


def extract_action(namespace):
    input_headers, pos, reader = custom_reader(namespace.report, ('status', 'filename', 'encoding'))

    selected_fields = namespace.select.split(',') if namespace.select else None
    selected_pos = [input_headers.index(h) for h in selected_fields] if selected_fields else None

    output_headers = (input_headers if not selected_pos else [input_headers[i] for i in selected_pos])
    output_headers += OUTPUT_ADDITIONAL_HEADERS

    if namespace.output is None:
        output_file = DummyTqdmFile(sys.stdout)
    else:
        output_file = open(namespace.output, 'w')

    output_writer = csv.writer(output_file)
    output_writer.writerow(output_headers)

    def files():
        for line in reader:
            status = int(line[pos.status]) if line[pos.status] else None

            if status is None or status >= 400:
                output_writer.writerow(line)
                loading_bar.update()
                continue

            path = join(namespace.input_directory, line[pos.filename])
            encoding = line[pos.encoding].strip() or 'utf-8'

            yield line, path, encoding

    loading_bar = tqdm(
        desc='Extracting content',
        total=namespace.total,
        dynamic_ncols=True,
        unit=' docs'
    )

    with Pool(namespace.processes) as pool:
        for error, line, content in pool.imap_unordered(worker, files()):
            loading_bar.update()

            if error is not None:
                message = ERROR_REPORTERS.get(type(error), repr(error))
                line.extend([message, ''])
                output_writer.writerow(line)
                continue

            line.extend(['', content])
            output_writer.writerow(line)

    output_file.close()



# =============================================================================
# Minet Facebook Mining Function
# =============================================================================
#
# A function fetching number of Facebook shares of a given url.
#
import re
from bs4 import BeautifulSoup
from urllib.parse import quote
import urllib3
urllib3.disable_warnings()
from ural import strip_protocol


def format_button_result(string):
    string = re.sub(r'\s', '', string)
    string = re.sub(r',', '.', string)
    try:
        digits = string.split('.')
        if len(digits) > 1:
            if digits[1][-1] == 'K':
                result = int(digits[0])*1000 + int(digits[1][:-1])*100
            elif digits[1][-1] == 'M':
                result = int(digits[0])*1000000 + int(digits[1][:-1])*100000
        elif len(digits) == 1:
            if digits[0][-1] == 'K':
                result = int(digits[0][:-1])*1000
            elif digits[0][-1] == 'M':
                result = int(digits[0][:-1])*1000000
            else:
                result = int(digits[0])
    except Exception as e:
        print(e)
        result = None

    return(result)


def fetch_share_count(page_url, detailed=False):
    """
    Function fetching the (approximate) number of Facebook shares for the given url.

    Args:
        page_url (str): Target URL as a string.

    Returns:
        dict: The HTML of the page.
    """
    page_url = strip_protocol(page_url)
    user_agent = {
        'user-agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:65.0) Gecko/20100101 Firefox/65.0'}
    pool = urllib3.PoolManager(headers=user_agent)
    try:
        http_button_request = "https://www.facebook.com/plugins/like.php?href=" + \
            quote("http://" + page_url) + "&layout=box_count"
        https_button_request = "https://www.facebook.com/plugins/like.php?href=" + \
            quote("https://" + page_url) + "&layout=box_count"
        http_button_html = pool.request(
            'GET', http_button_request).data.decode('utf-8')
        http_button_soup = BeautifulSoup(http_button_html, 'html.parser')
        http_button_result = http_button_soup.find(
            'span', attrs={"id": u"u_0_0"}).get_text()

        https_button_html = pool.request(
            'GET', https_button_request).data.decode('utf-8')
        https_button_soup = BeautifulSoup(https_button_html, 'html.parser')
        https_button_result = https_button_soup.find(
            'span', attrs={"id": u"u_0_0"}).get_text()
    except Exception as e:
        print(e)
        http_button_result = None
        https_button_result = None
    http_button_result = format_button_result(http_button_result)
    https_button_result = format_button_result(https_button_result)
    if detailed:
        return {"http": http_button_result, "https": https_button_result}
    else:
        # Seemingly working heuristic
        return max(http_button_result, https_button_result)


def fetch_share_counts(urls):
    for url in urls:
        yield fetch_share_count(url)



# =============================================================================
# Minet Contiguous Range Set Collection
# =============================================================================
#
# Simplistic implementation of a contiguous range set in python relying on a
# sorted list of contiguous intervals. It is very useful to represent a set of
# dense intervals using very little memory & is used across the task-resuming
# schemes of the CLI tool.
#
# It takes a lot of inspiration from the inversion list data structure.
#
from bisect import bisect_left


class ContiguousRangeSet(object):
    """
    Class representing contiguous ranges of integers as a sorted list of
    intervals.
    """

    def __init__(self):
        # NOTE: replace this by `blist` if not performany enough
        self.intervals = []
        self.current_stateful_interval = 0

    def add(self, point):
        """
        Method adding a single point to the set. We assume that the point
        cannot yet be in the set.
        """

        interval = (point, point)
        N = len(self.intervals)

        # Set is empty
        if N == 0:
            self.intervals.append(interval)
            return

        # Finding insertion point
        index = bisect_left(self.intervals, interval)

        if index >= N:
            last_interval = self.intervals[-1]

            if point == last_interval[1] + 1:
                self.intervals[-1] = (last_interval[0], point)
                return

            self.intervals.append(interval)
            return

        matched_interval = self.intervals[index]
        previous_interval = self.intervals[index - 1] if index != 0 else None

        if point == matched_interval[0] - 1:

            self.intervals[index] = (point, matched_interval[1])

            if previous_interval is not None:

                if point == previous_interval[1] + 1:
                    self.intervals[index] = (previous_interval[0], matched_interval[1])
                    self.intervals.pop(index - 1)

            return

        elif previous_interval is not None and point == previous_interval[1] + 1:

            self.intervals[index - 1] = (previous_interval[0], point)
            return

        if point < matched_interval[0]:
            self.intervals.insert(index, interval)
            return

    def stateful_contains(self, point):
        """
        Method returning whether the given point is found in the set. This
        method is stateful and assumes we will ask for points in a monotonic
        increasing order.
        """

        N = len(self.intervals)

        if N == 0 or self.current_stateful_interval >= N:
            return False

        I = self.intervals[self.current_stateful_interval]

        if point < I[0]:
            return False

        if point > I[1]:
            self.current_stateful_interval += 1
            return False

        return True




# =============================================================================
# Minet Scrape Function
# =============================================================================
#
# Function interpreting minet's scraping JSON descriptive DSL format to
# actually scrape content off of HTML content.
#
import re
from bs4 import BeautifulSoup


def extract_value(element, spec):

    # No specs -> returning text
    if spec is None:
        return element.get_text()

    # If spec is a string, we want an attribute
    if isinstance(spec, str):
        return element.get(spec)

    # Dictionary?
    fields = spec.get('fields')

    if fields is not None:
        o = {}

        for k, s in fields.items():
            v = extract_value(element, s)

            if v is not None:
                o[k] = v

        return o

    # Subselection
    sel = spec.get('sel')
    elements = [element]

    if sel is not None:
        elements = element.select(sel)

        if not elements:
            return spec.get('default')

        element = elements[0]

    value = element.get_text()

    # Retrieving attributes
    attr = spec.get('attr')

    if attr is not None:
        value = element.get(attr)

    # Retrieving raw text or html etc.
    # TODO...

    # Eval?
    expression = spec.get('eval')

    if expression is not None:
        return eval(expression, None, {
            're': re,
            'element': element,
            'elements': elements,
            'value': value
        })

    return value


def scrape_from_soup(soup, specs):
    item_specs = specs.get('item')

    iterator = specs.get('iterator')

    if iterator is None:
        elements = [soup]
    else:
        elements = soup.select(iterator)

    for element in elements:
        yield extract_value(element, item_specs)


def scrape(html, specs):
    soup = BeautifulSoup(html, 'lxml')

    return scrape_from_soup(soup, specs)


def headers_from_definition(specs):
    item = specs.get('item')
    fields = item.get('fields') if isinstance(item, dict) else None

    if fields is None or isinstance(item, str):
        return ['value']

    return list(item['fields'].keys())



# =============================================================================
# Minet Utils
# =============================================================================
#
# Miscellaneous helper function used throughout the library.
#
import re
import chardet
import cgi

# Handy regexes
CHARSET_RE = re.compile(rb'<meta.*?charset=["\']*(.+?)["\'>]', flags=re.I)
PRAGMA_RE = re.compile(rb'<meta.*?content=["\']*;?charset=(.+?)["\'>]', flags=re.I)
XML_RE = re.compile(rb'^<\?xml.*?encoding=["\']*(.+?)["\'>]')

# Constants
CHARDET_CONFIDENCE_THRESHOLD = 0.9


def guess_encoding(response, data, is_xml=False, use_chardet=False):
    """
    Function taking an urllib3 response object and attempting to guess its
    encoding.
    """
    content_type_header = response.getheader('content-type')

    if content_type_header is not None:
        parsed_header = cgi.parse_header(content_type_header)

        if len(parsed_header) > 1:
            charset = parsed_header[1].get('charset')

            if charset is not None:
                return charset.lower()

    # TODO: use re.search to go faster!
    if is_xml:
        matches = re.findall(CHARSET_RE, data)

        if len(matches) == 0:
            matches = re.findall(PRAGMA_RE, data)

        if len(matches) == 0:
            matches = re.findall(XML_RE, data)

        # NOTE: here we are returning the last one, but we could also use
        # frequency at the expense of performance
        if len(matches) != 0:
            return matches[-1].lower().decode()

    if use_chardet:
        chardet_result = chardet.detect(data)

        if chardet_result['confidence'] >= CHARDET_CONFIDENCE_THRESHOLD:
            return chardet_result['encoding'].lower()

    return None



# =============================================================================
# Minet CrowdTangle CLI Utils
# =============================================================================
#
# Miscellaneous generic functions used throughout the CrowdTangle actions.
#
import csv
import sys
import json
import time
import urllib3
import certifi
from tqdm import tqdm
from ural import get_domain_name, normalize_url

from minet.cli.utils import print_err
from minet.cli.crowdtangle.constants import CROWDTANGLE_DEFAULT_WAIT_TIME

URL_REPORT_HEADERS = [
    'post_ct_id',
    'post_id',
    'account_ct_id',
    'account_name',
    'account_handle',
    'original_url',
    'url',
    'normalized_url',
    'domain_name'
]


def format_url_for_csv(url_item, post):
    account = post['account']

    url = url_item['expanded']

    return [
        post['id'],
        post.get('platformId', ''),
        account['id'],
        account['name'],
        account.get('handle', ''),
        url_item['original'],
        url,
        normalize_url(url, strip_trailing_slash=True),
        get_domain_name(url)
    ]


def create_paginated_action(url_forge, csv_headers, csv_formatter,
                            item_name, item_key):

    def action(namespace, output_file):
        http = urllib3.PoolManager(
            cert_reqs='CERT_REQUIRED',
            ca_certs=certifi.where(),
        )

        url_report_writer = None

        if namespace.url_report:
            url_report_writer = csv.writer(namespace.url_report)
            url_report_writer.writerow(URL_REPORT_HEADERS)

        N = 0
        url = url_forge(namespace)

        has_limit = bool(namespace.limit)

        print_err('Using the following starting url:')
        print_err(url)
        print_err()

        # Loading bar
        loading_bar = tqdm(
            desc='Fetching %s' % item_name,
            dynamic_ncols=True,
            unit=' %s' % item_name,
            total=namespace.limit
        )

        if namespace.format == 'csv':
            writer = csv.writer(output_file)
            writer.writerow(csv_headers(namespace) if callable(csv_headers) else csv_headers)

        while True:
            result = http.request('GET', url)

            if result.status == 401:
                loading_bar.close()

                print_err('Your API token is invalid.')
                print_err('Check that you indicated a valid one using the `--token` argument.')
                sys.exit(1)

            if result.status >= 400:
                loading_bar.close()

                print_err(result.data, result.status)
                sys.exit(1)

            try:
                data = json.loads(result.data)['result']
            except:
                loading_bar.close()
                print_err('Misformatted JSON result.')
                sys.exit(1)

            if item_key not in data or len(data[item_key]) == 0:
                break

            items = data[item_key]
            enough_to_stop = False

            for item in items:
                N += 1
                loading_bar.update()

                if namespace.format == 'jsonl':
                    output_file.write(json.dumps(item, ensure_ascii=False) + '\n')
                else:
                    writer.writerow(csv_formatter(namespace, item))

                if url_report_writer:
                    urls = item.get('expandedLinks')

                    if urls:
                        for url_item in urls:
                            url_report_writer.writerow(format_url_for_csv(url_item, item))

                if has_limit and N >= namespace.limit:
                    enough_to_stop = True
                    break

            # NOTE: I wish I had labeled loops in python...
            if enough_to_stop:
                loading_bar.close()
                print_err('The indicated limit of %s was reached.' % item_name)
                break

            # Pagination
            # NOTE: we could adjust the `count` GET param but I am lazy
            pagination = data['pagination']

            if 'nextPage' not in pagination:
                loading_bar.close()
                print_err('We reached the end of pagination.')
                break

            url = pagination['nextPage']

            # Waiting a bit to respect the 6 reqs/min limit
            time.sleep(CROWDTANGLE_DEFAULT_WAIT_TIME)

        loading_bar.close()


    return action



#!/usr/bin/env python
# =============================================================================
# Minet CLI Endpoint
# =============================================================================
#
# CLI enpoint of the Minet library.
#
import csv
import sys
import shutil
from textwrap import dedent
from argparse import ArgumentParser, FileType, RawTextHelpFormatter

from minet.cli.defaults import DEFAULT_CONTENT_FOLDER
from minet.cli.utils import BooleanAction

from minet.cli.crowdtangle.constants import CROWDTANGLE_SORT_TYPES

SUBPARSERS = {}

terminal_size = shutil.get_terminal_size()
csv.field_size_limit(sys.maxsize)

def custom_formatter(prog):
    return RawTextHelpFormatter(
        prog,
        max_help_position=50,
        width=terminal_size.columns,
    )


def main():
    parser = ArgumentParser(prog='minet')
    subparsers = parser.add_subparsers(
        help='Action to execute', title='actions', dest='action')

    # Fetch action subparser
    # -------------------------------------------------------------------------
    fetch_description = dedent(
        '''
        Minet Fetch Command
        ===================

        Use multiple threads to fetch batches of urls from a CSV file. The
        command outputs a CSV report with additional metadata about the
        HTTP calls and will generally write the retrieved files in a folder
        given by the user.
        '''
    )

    fetch_epilog = dedent(
        '''
        examples:

        . Fetching a batch of url from existing CSV file:
            `minet fetch url_column file.csv > report.csv`

        . CSV input from stdin:
            `xsv select url_column file.csv | minet fetch url_column > report.csv`

        . Fetching a single url, useful to pipe into `minet scrape`:
            `minet fetch http://google.com | minet scrape ./scrape.json > scraped.csv`
        '''
    )

    fetch_subparser = subparsers.add_parser(
        'fetch',
        description=fetch_description,
        epilog=fetch_epilog,
        formatter_class=custom_formatter
    )

    fetch_subparser.add_argument(
        'column',
        help='Column of the CSV file containing urls to fetch.'
    )

    fetch_subparser.add_argument(
        'file',
        help='CSV file containing the urls to fetch.',
        type=FileType('r'),
        default=sys.stdin,
        nargs='?'
    )

    fetch_subparser.add_argument(
        '--contents-in-report',
        help='Whether to include retrieved contents, e.g. html, directly in the report\nand avoid writing them in a separate folder. This requires to standardize\nencoding and won\'t work on binary formats.',
        action='store_true'
    )
    fetch_subparser.add_argument(
        '-d', '--output-dir',
        help='Directory where the fetched files will be written. Defaults to "%s".' % DEFAULT_CONTENT_FOLDER,
        default=DEFAULT_CONTENT_FOLDER
    )
    fetch_subparser.add_argument(
        '-f', '--filename',
        help='Name of the column used to build retrieved file names. Defaults to an uuid v4 with correct extension.'
    )
    fetch_subparser.add_argument(
        '--filename-template',
        help='A template for the name of the fetched files.'
    )
    fetch_subparser.add_argument(
        '-g', '--grab-cookies',
        help='Whether to attempt to grab cookies from your computer\'s chrome browser.',
        action='store_true'
    )
    fetch_subparser.add_argument(
        '--standardize-encoding',
        help='Whether to systematically convert retrieved text to UTF-8.',
        action='store_true'
    )
    fetch_subparser.add_argument(
        '-o', '--output',
        help='Path to the output report file. By default, the report will be printed to stdout.'
    )
    fetch_subparser.add_argument(
        '-s', '--select',
        help='Columns to include in report (separated by `,`).'
    )
    fetch_subparser.add_argument(
        '-t', '--threads',
        help='Number of threads to use. Defaults to 25.',
        type=int,
        default=25
    )
    fetch_subparser.add_argument(
        '--throttle',
        help='Time to wait - in seconds - between 2 calls to the same domain. Defaults to 0.2.',
        type=float,
        default=0.2
    )
    fetch_subparser.add_argument(
        '--total',
        help='Total number of lines in CSV file. Necessary if you want to display a finite progress indicator.',
        type=int
    )
    fetch_subparser.add_argument(
        '--url-template',
        help='A template for the urls to fetch. Handy e.g. if you need to build urls from ids etc.'
    )

    # TODO: lru_cache, normalize urls, print current urls? print end report?

    SUBPARSERS['fetch'] = fetch_subparser

    # Crowdtangle actions subparser
    # -------------------------------------------------------------------------
    crowdtangle_description = dedent(
        '''
        Minet CrowdTangle Command
        =========================

        Gather data from the CrowdTangle APIs easily and efficiently.
        '''
    )

    crowdtangle_subparser = subparsers.add_parser(
        'ct',
        description=crowdtangle_description,
        formatter_class=custom_formatter
    )

    crowdtangle_subparser_subparsers = crowdtangle_subparser.add_subparsers(
        help='Action to perform using the CrowdTangle API.',
        title='actions',
        dest='ct_action'
    )

    def common_ct_arguments(sub):
        sub.add_argument(
            '-o', '--output',
            help='Path to the output file. By default, everything will be printed to stdout.'
        )
        sub.add_argument(
            '-t', '--token',
            help='CrowdTangle dashboard API token.'
        )

    common_ct_arguments(crowdtangle_subparser)

    # Crowdtangle leaderboard `ct leaderboard`
    crowdtangle_leaderboard_subparser = crowdtangle_subparser_subparsers.add_parser(
        'leaderboard',
        description=dedent(
            '''
            Minet CrowdTangle Leaderboard Command
            =====================================

            Gather information and aggregated stats about pages and groups of
            the designated dashboard (indicated by a given token).
            '''
        ),
        epilog=dedent(
        '''
            examples:

            . Fetching accounts statistics for every account in your dashboard:
                `minet ct leaderboard --token YOUR_TOKEN > accounts-stats.csv`
            '''
        ),
        formatter_class=custom_formatter
    )

    common_ct_arguments(crowdtangle_leaderboard_subparser)

    crowdtangle_leaderboard_subparser.add_argument(
        '--no-breakdown',
        help='Whether to skip statistics breakdown by post type in the CSV output.',
        dest='breakdown',
        action=BooleanAction,
        default=True
    )
    crowdtangle_leaderboard_subparser.add_argument(
        '-f', '--format',
        help='Output format. Defaults to `csv`.',
        choices=['csv', 'jsonl'],
        default='csv'
    )
    crowdtangle_leaderboard_subparser.add_argument(
        '-l', '--limit',
        help='Maximum number of posts to retrieve. Will fetch every post by default.',
        type=int
    )
    crowdtangle_leaderboard_subparser.add_argument(
        '--list-id',
        help='Optional list id from which to retrieve accounts.'
    )

    # Crowdtangle lists `ct lists`
    crowdtangle_lists_subparser = crowdtangle_subparser_subparsers.add_parser(
        'lists',
        description=dedent(
            '''
            Minet CrowdTangle Lists Command
            ===============================

            Retrieve the lists from a CrowdTangle dashboard (indicated by a
            given token).
            '''
        ),
        epilog=dedent(
        '''
            examples:

            . Fetching a dashboard's lists:
                `minet ct lists --token YOUR_TOKEN > lists.csv`
            '''
        ),
        formatter_class=custom_formatter
    )

    common_ct_arguments(crowdtangle_lists_subparser)

    # Crowdtangle posts `ct posts`
    crowdtangle_posts_subparser = crowdtangle_subparser_subparsers.add_parser(
        'posts',
        description=dedent(
            '''
            Minet CrowdTangle Posts Command
            ===============================

            Gather post data from the designated dashboard (indicated by
            a given token).
            '''
        ),
        epilog=dedent(
        '''
            examples:

            . Fetching the 500 most latest posts from a dashboard:
                `minet ct posts --token YOUR_TOKEN --limit 500 > latest-posts.csv`
            '''
        ),
        formatter_class=custom_formatter
    )

    common_ct_arguments(crowdtangle_posts_subparser)

    crowdtangle_posts_subparser.add_argument(
        '--end-date',
        help='The latest date at which a post could be posted (UTC!).'
    )
    crowdtangle_posts_subparser.add_argument(
        '-f', '--format',
        help='Output format. Defaults to `csv`.',
        choices=['csv', 'jsonl'],
        default='csv'
    )
    crowdtangle_posts_subparser.add_argument(
        '--language',
        help='Language of posts to retrieve.'
    )
    crowdtangle_posts_subparser.add_argument(
        '-l', '--limit',
        help='Maximum number of posts to retrieve. Will fetch every post by default.',
        type=int
    )
    crowdtangle_posts_subparser.add_argument(
        '--list-ids',
        help='Ids of the lists from which to retrieve posts, separated by commas.'
    )
    crowdtangle_posts_subparser.add_argument(
        '--sort-by',
        help='The order in which to retrieve posts. Defaults to `date`.',
        choices=CROWDTANGLE_SORT_TYPES,
        default='date'
    )
    crowdtangle_posts_subparser.add_argument(
        '--start-date',
        help='The earliest date at which a post could be posted (UTC!).'
    )
    crowdtangle_posts_subparser.add_argument(
        '--url-report',
        help='Path to an optional report file to write about urls found in posts.',
        type=FileType('w')
    )

    # Crowdtangle search `ct search`
    crowdtangle_search_subparser = crowdtangle_subparser_subparsers.add_parser(
        'search',
        description=dedent(
            '''
            Minet CrowdTangle Search Command
            ================================

            Search posts on the whole CrowdTangle platform.
            '''
        ),
        epilog=dedent(
        '''
            examples:

            . Fetching a dashboard's lists:
                `minet ct search --token YOUR_TOKEN > posts.csv`
            '''
        ),
        formatter_class=custom_formatter
    )

    common_ct_arguments(crowdtangle_search_subparser)

    crowdtangle_search_subparser.add_argument(
        'terms',
        help='The search query term or terms.'
    )

    crowdtangle_search_subparser.add_argument(
        '--end-date',
        help='The latest date at which a post could be posted (UTC!).'
    )
    crowdtangle_search_subparser.add_argument(
        '-f', '--format',
        help='Output format. Defaults to `csv`.',
        choices=['csv', 'jsonl'],
        default='csv'
    )
    crowdtangle_search_subparser.add_argument(
        '-l', '--limit',
        help='Maximum number of posts to retrieve. Will fetch every post by default.',
        type=int
    )
    crowdtangle_search_subparser.add_argument(
        '--offset',
        help='Count offset.',
        type=int
    )
    crowdtangle_search_subparser.add_argument(
        '-p', '--platforms',
        help='The platforms, separated by comma from which to retrieve posts.'
    )
    crowdtangle_search_subparser.add_argument(
        '--sort-by',
        help='The order in which to retrieve posts. Defaults to `date`.',
        choices=CROWDTANGLE_SORT_TYPES,
        default='date'
    )
    crowdtangle_search_subparser.add_argument(
        '--start-date',
        help='The earliest date at which a post could be posted (UTC!).'
    )
    crowdtangle_search_subparser.add_argument(
        '--types',
        help='Types of post to include, separated by comma.'
    )
    crowdtangle_search_subparser.add_argument(
        '--url-report',
        help='Path to an optional report file to write about urls found in posts.',
        type=FileType('w')
    )

    SUBPARSERS['ct'] = crowdtangle_subparser

    # Extract action subparser
    # -------------------------------------------------------------------------
    extract_description = dedent(
        '''
        Minet Extract Command
        =====================

        Use multiple processes to extract raw text from a batch of HTML files.
        This command can either work on a `minet fetch` report or on a bunch
        of files. It will output an augmented report with the extracted text.
        '''
    )

    extract_epilog = dedent(
        '''
        examples:

        . Extracting raw text from a `minet fetch` report:
            `minet extract report.csv > extracted.csv`

        . Working on a report from stdin:
            `minet fetch url_column file.csv | minet extract > extracted.csv`

        . Extracting raw text from a bunch of files:
            `minet extract --glob "./content/*.html" > extracted.csv`
        '''
    )

    extract_subparser = subparsers.add_parser(
        'extract',
        description=extract_description,
        epilog=extract_epilog,
        formatter_class=custom_formatter
    )

    extract_subparser.add_argument(
        'report',
        help='Input CSV fetch action report file.',
        type=FileType('r'),
        default=sys.stdin,
        nargs='?'
    )

    extract_subparser.add_argument(
        '-e', '--extractor',
        help='Extraction engine to use. Defaults to `dragnet`.',
        choices=['dragnet', 'html2text']
    )
    extract_subparser.add_argument(
        '-i', '--input-directory',
        help='Directory where the HTML files are stored. Defaults to "%s".' % DEFAULT_CONTENT_FOLDER,
        default=DEFAULT_CONTENT_FOLDER
    )
    extract_subparser.add_argument(
        '-o', '--output',
        help='Path to the output report file. By default, the report will be printed to stdout.'
    )
    extract_subparser.add_argument(
        '-p', '--processes',
        help='Number of processes to use. Defaults to 4.',
        type=int,
        default=4
    )
    extract_subparser.add_argument(
        '-s', '--select',
        help='Columns to include in report (separated by `,`).'
    )
    extract_subparser.add_argument(
        '--total',
        help='Total number of HTML documents. Necessary if you want to display a finite progress indicator.',
        type=int
    )

    SUBPARSERS['extract'] = extract_subparser

    # Scrape action subparser
    # -------------------------------------------------------------------------
    scrape_description = dedent(
        '''
        Minet Scrape Command
        ====================

        Use multiple processes to scrape data from a batch of HTML files.
        This command can either work on a `minet fetch` report or on a bunch
        of files. It will output the scraped items.
        '''
    )

    scrape_epilog = dedent(
        '''
        examples:

        . Scraping item from a `minet fetch` report:
            `minet scrape scraper.json report.csv > scraped.csv`

        . Working on a report from stdin:
            `minet fetch url_column file.csv | minet fetch scraper.json > scraped.csv`

        . Scraping items from a bunch of files:
            `minet scrape scraper.json --glob "./content/*.html" > scraped.csv`
        '''
    )

    scrape_subparser = subparsers.add_parser(
        'scrape',
        description=scrape_description,
        epilog=scrape_epilog,
        formatter_class=custom_formatter
    )

    scrape_subparser.add_argument(
        'scraper',
        help='Path to a scraper definition file.',
        type=FileType('r')
    )

    scrape_subparser.add_argument(
        'report',
        help='Input CSV fetch action report file.',
        type=FileType('r'),
        default=sys.stdin,
        nargs='?'
    )

    scrape_subparser.add_argument(
        '-i', '--input-directory',
        help='Directory where the HTML files are stored. Defaults to "%s".' % DEFAULT_CONTENT_FOLDER,
        default=DEFAULT_CONTENT_FOLDER
    )
    scrape_subparser.add_argument(
        '-o', '--output',
        help='Path to the output report file. By default, the report will be printed to stdout.'
    )
    scrape_subparser.add_argument(
        '-p', '--processes',
        help='Number of processes to use. Defaults to 4.',
        type=int,
        default=4
    )
    scrape_subparser.add_argument(
        '--total',
        help='Total number of HTML documents. Necessary if you want to display a finite progress indicator.',
        type=int
    )

    SUBPARSERS['scrape'] = scrape_subparser

    # Help subparser
    # -------------------------------------------------------------------------
    help_suparser = subparsers.add_parser('help')
    help_suparser.add_argument('subcommand', help='Name of the subcommand', nargs='?')
    SUBPARSERS['help'] = help_suparser

    args = parser.parse_args()

    # TODO: handle sub commands?
    if args.action == 'help':
        target_subparser = SUBPARSERS.get(args.subcommand)

        if target_subparser is None:
            parser.print_help()
        else:
            target_subparser.print_help()

    elif args.action == 'fetch':
        from minet.cli.fetch import fetch_action
        fetch_action(args)

    elif args.action == 'ct':
        from minet.cli.crowdtangle import crowdtangle_action
        crowdtangle_action(args)

    elif args.action == 'extract':
        try:
            import dragnet
        except:
            print('The `dragnet` library is not installed. The `extract` command won\'t work.')
            print('To install it correctly, run the following commands in order:')
            print()
            print('  pip install lxml numpy Cython')
            print('  pip install dragnet')
            sys.exit(1)

        from minet.cli.extract import extract_action
        extract_action(args)

    elif args.action == 'scrape':
        from minet.cli.scrape import scrape_action
        scrape_action(args)

    else:
        parser.print_help()

if __name__ == '__main__':
    main()



# =============================================================================
# Minet Fetch CLI Action
# =============================================================================
#
# Action reading an input CSV file line by line and fetching the urls found
# in the given column. This is done in a respectful multithreaded fashion to
# optimize both running time & memory.
#
import os
import csv
import sys
import json
import certifi
import mimetypes
from io import StringIO
from os.path import join
from collections import Counter
from urllib3 import PoolManager, Timeout
from tqdm import tqdm
from quenouille import imap_unordered
from uuid import uuid4
from ural import ensure_protocol, is_url, get_domain_name
from pycookiecheat import chrome_cookies

from urllib3.exceptions import (
    HTTPError,
    ClosedPoolError,
    ConnectTimeoutError,
    MaxRetryError,
    ReadTimeoutError,
    ResponseError
)

from minet.utils import guess_encoding
from minet.cli.utils import custom_reader, DummyTqdmFile

mimetypes.init()

OUTPUT_ADDITIONAL_HEADERS = ['line', 'status', 'error', 'filename', 'encoding']

# TODO: make this an option!
SPOOFED_UA = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.103 Safari/537.36'

def max_retry_error_reporter(error):
    if isinstance(error, (ConnectTimeoutError, ReadTimeoutError)):
        return 'timeout'

    if isinstance(error.reason, ResponseError) and 'redirect' in repr(error.reason):
        return 'too-many-redirects'

    return 'max-retries-exceeded'

ERROR_REPORTERS = {
    MaxRetryError: max_retry_error_reporter,
    UnicodeEncodeError: 'headers-encoding'
}


def fetch(http, url, cookies=None):

    # Formatting headers
    headers = {
        'User-Agent': SPOOFED_UA
    }

    # Formatting cookies header
    if cookies is not None:
        headers['Cookie'] = '; '.join('%s=%s' % r for r in cookies.items())

    # TODO: probably need to do redirects myself to avoid ClosedPoolError
    try:
        r = http.request(
            'GET',
            url,
            headers=headers
        )

        return None, r

    except ClosedPoolError:

        # TODO: this is a clunky workaround
        return fetch(http, url)

    # TODO: when urllib3 updates and release #1487, we'll need to change that
    except (HTTPError, UnicodeEncodeError) as e:
        return e, None


def worker(job):
    """
    Function using the urllib3 http to actually fetch our contents from the web.
    """
    namespace, http, line, url = job

    url = ensure_protocol(url)
    cookies = None

    # Should we grab cookies?
    if namespace.grab_cookies:
        cookies = chrome_cookies(url)

        if len(cookies) == 0:
            cookies = None

    error, result = fetch(http, url, cookies=cookies)

    if error:
        return error, url, line, result, None, None

    # Forcing urllib3 to read data in thread
    data = result.data

    # Solving mime type
    (mimetype, _) = mimetypes.guess_type(url)

    if mimetype is None:
        mimetype = 'text/html'

    exts = mimetypes.guess_all_extensions(mimetype)

    if not exts:
        ext = '.html'
    else:
        ext = max(exts, key=len)

    # Solving encoding
    is_xml = ext == '.html' or ext == '.xml'

    encoding = guess_encoding(result, data, is_xml=is_xml, use_chardet=True)

    info = {
        'mime': mimetype,
        'ext': ext,
        'encoding': encoding
    }

    return error, url, line, result, data, info


def fetch_action(namespace):

    # Do we need to fetch only a single url?
    if namespace.file is sys.stdin and is_url(namespace.column):
        namespace.file = StringIO('url\n%s' % namespace.column)
        namespace.column = 'url'

    input_headers, pos, reader = custom_reader(namespace.file, namespace.column)
    filename_pos = input_headers.index(namespace.filename) if namespace.filename else None

    selected_fields = namespace.select.split(',') if namespace.select else None
    selected_pos = [input_headers.index(h) for h in selected_fields] if selected_fields else None

    # First we need to create the relevant directory
    if not namespace.contents_in_report:
        os.makedirs(namespace.output_dir, exist_ok=True)

    # Loading bar
    loading_bar = tqdm(
        desc='Fetching pages',
        total=namespace.total,
        dynamic_ncols=True,
        unit=' urls'
    )

    # Reading output
    output_headers = (input_headers if not selected_pos else [input_headers[i] for i in selected_pos])
    output_headers += OUTPUT_ADDITIONAL_HEADERS

    if namespace.contents_in_report:
        output_headers.append('raw_content')

    if namespace.output is None:
        output_file = DummyTqdmFile(sys.stdout)
    else:
        output_file = open(namespace.output, 'w')

    output_writer = csv.writer(output_file)
    output_writer.writerow(output_headers)

    # Creating the http pool manager
    http = PoolManager(
        cert_reqs='CERT_REQUIRED',
        ca_certs=certifi.where(),
        num_pools=namespace.threads * 2,
        maxsize=1, # NOTE: should be the same as group_parallelism,
        timeout=Timeout(connect=2.0, read=7.0)
    )

    # Generator yielding urls to fetch
    def payloads():
        for line in reader:
            url = line[pos]

            if not url:

                # TODO: write report line all the same!
                loading_bar.update()
                continue

            # Url templating
            if namespace.url_template:
                url = namespace.url_template.format(value=url)
            else:
                url = url.strip()

            yield (namespace, http, line, url)

    # Streaming the file and fetching the url using multiple threads
    multithreaded_iterator = imap_unordered(
        payloads(),
        worker,
        namespace.threads,
        group=lambda job: get_domain_name(job[3]),
        group_parallelism=1,
        group_buffer_size=25,
        group_throttle=namespace.throttle
    )

    errors = 0
    status_codes = Counter()

    for i, (error, url, line, result, data, info) in enumerate(multithreaded_iterator):

        content_write_flag = 'wb'

        # Updating stats
        if error is not None:
            errors += 1
        else:
            if result.status >= 400:
                status_codes[result.status] += 1

        postfix = {'errors': errors}

        for code, count in status_codes.most_common(1):
            postfix[str(code)] = count

        loading_bar.set_postfix(**postfix)
        loading_bar.update()

        # No error
        if error is None:

            filename = None

            # Building filename
            if filename_pos is not None:
                if namespace.filename_template:
                    filename = namespace.filename_template.format(value=line[filename_pos])
                else:
                    filename = line[filename_pos] + info['ext']
            else:
                # NOTE: it would be nice to have an id that can be sorted by time
                filename = str(uuid4()) + info['ext']

            # Standardize encoding?
            encoding = info['encoding']

            if namespace.standardize_encoding or namespace.contents_in_report:
                if encoding is None or encoding != 'utf-8' or namespace.contents_in_report:
                    data = data.decode(encoding, errors='replace')
                    encoding = 'utf-8'
                    content_write_flag = 'w'

            # Writing file on disk
            if not namespace.contents_in_report:
                with open(join(namespace.output_dir, filename), content_write_flag) as f:
                    f.write(data)

            # Reporting in output
            if selected_pos:
                line = [line[i] for i in selected_pos]

            line.extend([i, result.status, '', filename, encoding or ''])

            if namespace.contents_in_report:
                line.append(data)

            output_writer.writerow(line)

        # Handling potential errors
        else:
            reporter = ERROR_REPORTERS.get(type(error), repr)

            error_code = reporter(error) if callable(reporter) else reporter

            # Reporting in output
            if selected_pos:
                line = [line[i] for i in selected_pos]

            line.extend([i, '', error_code, '', ''])
            output_writer.writerow(line)

    # Closing files
    if namespace.output is not None:
        output_file.close()



from setuptools import setup, find_packages

with open('./README.md', 'r') as f:
    long_description = f.read()

setup(name='minet',
      version='0.6.0',
      description='A webmining CLI tool & library for python.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='http://github.com/medialab/minet',
      license='MIT',
      author='Jules Farjas, Guillaume Plique',
      keywords='webmining',
      python_requires='>=3',
      packages=find_packages(exclude=['ftest', 'test']),
      install_requires=[
        'beautifulsoup4>=4.7.1',
        'chardet>=3.0.4',
        'cython>=0.29.4',
        'lxml>=4.3.0',
        'numpy>=1.16.1',
        'pycookiecheat>=0.4.3',
        'quenouille>=0.3.0',
        'tqdm>=4.31.1',
        'ural>=0.7.0',
        'urllib3[secure]>=1.24.2',
        # 'dragnet>=2.0.3'
      ],
      entry_points={
        'console_scripts': ['minet=minet.cli.__main__:main']
      },
      zip_safe=True)



# =============================================================================
# Minet CrowdTangle Posts CLI Action
# =============================================================================
#
# Logic of the `ct posts` action.
#
from minet.cli.crowdtangle.utils import create_paginated_action
from minet.cli.crowdtangle.constants import CROWDTANGLE_POST_TYPES

URL_TEMPLATE = 'https://api.crowdtangle.com/leaderboard?count=100&token=%s'


def forge_leaderboard_url(namespace):
    base_url = URL_TEMPLATE % namespace.token

    if namespace.list_id:
        base_url += '&listId=%s' % namespace.list_id

    return base_url


CSV_HEADERS = [
    'ct_id',
    'name',
    'handle',
    'profile_image',
    'subscriber_count',
    'url',
    'verified',
    'initial_subscriber_count',
    'final_subscriber_count',
    'subscriber_data_notes'
]

STATISTICS = [
    ('loveCount', 'love_count'),
    ('wowCount', 'wow_count'),
    ('thankfulCount', 'thankful_count'),
    ('interactionRate', 'interaction_rate'),
    ('likeCount', 'like_count'),
    ('hahaCount', 'haha_count'),
    ('commentCount', 'comment_count'),
    ('shareCount', 'share_count'),
    ('sadCount', 'sad_count'),
    ('angryCount', 'angry_count'),
    ('postCount', 'post_count'),
    ('totalInteractionCount', 'total_interaction_count'),
    ('totalVideoTimeMS', 'total_video_time_ms'),
    ('threePlusMinuteVideoCount', 'three_plus_minute_video_count')
]

# Summary keys
for _, substitute_key in STATISTICS:
    CSV_HEADERS.append(substitute_key)


def build_csv_headers(namespace):
    basic_headers = list(CSV_HEADERS)

    if not namespace.breakdown:
        return basic_headers

    for post_type in CROWDTANGLE_POST_TYPES:
        for _, substitute_key in STATISTICS:
            basic_headers.append('%s_%s' % (post_type, substitute_key))

    return basic_headers


def format_account_for_csv(namespace, item):
    account = item['account']
    subscriber_data = item['subscriberData']

    row = [
        account['id'],
        account['name'],
        account.get('handle', ''),
        account['profileImage'],
        account['subscriberCount'],
        account['url'],
        '1' if account['verified'] else '',
        subscriber_data['initialCount'],
        subscriber_data['finalCount'],
        subscriber_data.get('notes', '')
    ]

    summary = item['summary']

    for key, _ in STATISTICS:
        row.append(summary.get(key, ''))

    if not namespace.breakdown:
        return row

    breakdown = item['breakdown']

    for post_type in CROWDTANGLE_POST_TYPES:

        data = breakdown.get(post_type)

        for key, _ in STATISTICS:
            row.append(data.get(key, '') if data else '')

    return row


crowdtangle_leaderboard_action = create_paginated_action(
    url_forge=forge_leaderboard_url,
    csv_headers=build_csv_headers,
    csv_formatter=format_account_for_csv,
    item_name='accounts',
    item_key='accountStatistics'
)
